"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RelendAction = void 0;
const spl_token_1 = require("@solana/spl-token");
const web3_js_1 = require("@solana/web3.js");
const bignumber_js_1 = __importDefault(require("bignumber.js"));
const bn_js_1 = __importDefault(require("bn.js"));
const constants_1 = require("../core/constants");
const instructions_1 = require("../instructions");
const obligation_1 = require("../state/obligation");
const reserve_1 = require("../state/reserve");
const constants_2 = require("./constants");
const utils_1 = require("./utils");
// const API_ENDPOINT = "https://api.solend.fi"
const API_ENDPOINT = process.env.NEXT_PUBLIC_BACKEND_URL;
const SOL_PADDING_FOR_INTEREST = "1000000";
function getTokenInfo(symbol, solendInfo) {
    const tokenInfo = solendInfo.reserves.find((reserve) => reserve.liquidityToken.symbol === symbol);
    if (!tokenInfo) {
        throw new Error(`Could not find ${symbol} in ASSETS`);
    }
    return tokenInfo;
}
const PRICE_PAIRS = (0, constants_1.getPricePair)();
const REUSD_PRICE = (0, constants_1.getReUSDPricePair)();
const REVND_PRICE = constants_1.PRICE_PAIRS_MAINNET_TEST_REVND_RENGN[0];
const RENGN_PRICE = constants_1.PRICE_PAIRS_MAINNET_TEST_REVND_RENGN[1];
const WITHDRAW_BUFFER_SLOTS = 80;
class RelendAction {
    programId;
    connection;
    reserve;
    lendingMarket;
    publicKey;
    obligationAddress;
    obligationAccountInfo;
    userTokenAccountAddress;
    userCollateralAccountAddress;
    seed;
    symbol;
    positions;
    amount;
    hostAta;
    setupIxs;
    lendingIxs;
    cleanupIxs;
    preTxnIxs;
    postTxnIxs;
    depositReserves;
    borrowReserves;
    constructor(programId, connection, reserve, lendingMarket, publicKey, obligationAddress, obligationAccountInfo, userTokenAccountAddress, userCollateralAccountAddress, seed, symbol, positions, amount, depositReserves, borrowReserves, hostAta, setupIxs = []) {
        this.programId = programId;
        this.connection = connection;
        this.publicKey = publicKey;
        this.amount = new bn_js_1.default(amount);
        this.symbol = symbol;
        this.positions = positions;
        this.hostAta = hostAta;
        this.obligationAccountInfo = obligationAccountInfo;
        this.lendingMarket = lendingMarket;
        this.seed = seed;
        this.reserve = reserve;
        this.obligationAddress = obligationAddress;
        this.userTokenAccountAddress = userTokenAccountAddress;
        this.userCollateralAccountAddress = userCollateralAccountAddress;
        this.setupIxs = setupIxs;
        this.lendingIxs = [];
        this.cleanupIxs = [];
        this.preTxnIxs = [];
        this.postTxnIxs = [];
        this.depositReserves = depositReserves;
        this.borrowReserves = borrowReserves;
    }
    static async initialize(action, amount, symbol, publicKey, connection, environment = "production", lendingMarketAddress, hostAddress) {
        // const solendInfo = (await (
        //   await axios.get(
        //     `${API_ENDPOINT}/v1/markets/configs?scope=all&deployment=${environment}`
        //   )
        // ).data) as ConfigType;
        const relendInfo = (0, constants_1.getRelendInfo)();
        let lendingMarket;
        if (lendingMarketAddress) {
            lendingMarket = relendInfo.find((market) => market.address == lendingMarketAddress.toBase58());
            if (!lendingMarket) {
                throw `market address not found: ${lendingMarketAddress}`;
            }
        }
        else {
            lendingMarket = relendInfo.find((market) => market.isPrimary) ?? relendInfo[0];
        }
        const seed = lendingMarket.address.slice(0, 32);
        const programId = (0, constants_2.getProgramId)();
        const obligationAddress = await web3_js_1.PublicKey.createWithSeed(publicKey, seed, programId);
        const reserve = lendingMarket.reserves.find((res) => res.liquidityToken.symbol === symbol);
        if (!reserve) {
            throw new Error(`Could not find asset ${symbol} in reserves`);
        }
        const obligationAccountInfo = await connection.getAccountInfo(obligationAddress);
        let obligationDetails = null;
        const depositReserves = [];
        const borrowReserves = [];
        if (obligationAccountInfo) {
            obligationDetails = (0, obligation_1.parseObligation)(web3_js_1.PublicKey.default, obligationAccountInfo).info;
            obligationDetails.deposits.forEach((deposit) => {
                depositReserves.push(deposit.depositReserve);
            });
            obligationDetails.borrows.forEach((borrow) => {
                borrowReserves.push(borrow.borrowReserve);
            });
        }
        // Union of addresses
        const distinctReserveCount = [
            ...Array.from(new Set([...borrowReserves.map((e) => e.toBase58()), ...(action === "borrow" ? [reserve.address] : [])])),
        ].length +
            [
                ...Array.from(new Set([...depositReserves.map((e) => e.toBase58()), ...(action === "deposit" ? [reserve.address] : [])])),
            ].length;
        if (distinctReserveCount > constants_1.POSITION_LIMIT) {
            throw Error(`Obligation already has max number of positions: ${constants_1.POSITION_LIMIT}`);
        }
        const tokenInfo = getTokenInfo(symbol, lendingMarket);
        const userTokenAccountAddress = await spl_token_1.Token.getAssociatedTokenAddress(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, new web3_js_1.PublicKey(tokenInfo.liquidityToken.mint), publicKey, true);
        const userCollateralAccountAddress = await spl_token_1.Token.getAssociatedTokenAddress(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, new web3_js_1.PublicKey(reserve.collateralMintAddress), publicKey, true);
        // Host ATA
        let hostAta = undefined;
        const setupIxs = [];
        if (hostAddress) {
            const hostTokenAccountAddress = await spl_token_1.Token.getAssociatedTokenAddress(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, new web3_js_1.PublicKey(reserve.liquidityToken.mint), new web3_js_1.PublicKey(hostAddress));
            const hostTokenAccountInfo = await connection.getAccountInfo(hostTokenAccountAddress);
            if (!hostTokenAccountInfo) {
                setupIxs.push((0, utils_1.createAssociatedTokenAccountInstruction)(publicKey, hostTokenAccountAddress, hostAddress, new web3_js_1.PublicKey(reserve.liquidityToken.mint)));
            }
            hostAta = hostTokenAccountAddress;
        }
        return new RelendAction(programId, connection, reserve, lendingMarket, publicKey, obligationAddress, obligationDetails, userTokenAccountAddress, userCollateralAccountAddress, seed, symbol, distinctReserveCount, amount, depositReserves, borrowReserves, hostAta, setupIxs);
    }
    static async buildDepositTxns(connection, amount, symbol, publicKey, environment = "production", lendingMarketAddress) {
        const axn = await RelendAction.initialize("deposit", amount, symbol, publicKey, connection, environment, lendingMarketAddress);
        await axn.addSupportIxs("deposit");
        await axn.addDepositIx();
        return axn;
    }
    static async buildBorrowTxns(connection, amount, symbol, publicKey, environment = "production", hostAddress, lendingMarketAddress) {
        const axn = await RelendAction.initialize("borrow", amount, symbol, publicKey, connection, environment, lendingMarketAddress, hostAddress);
        await axn.addSupportIxs("borrow");
        await axn.addBorrowIx();
        return axn;
    }
    static async buildDepositReserveLiquidityTxns(connection, amount, symbol, publicKey, environment = "production", lendingMarketAddress) {
        const axn = await RelendAction.initialize("mint", amount, symbol, publicKey, connection, environment, lendingMarketAddress);
        await axn.addSupportIxs("mint");
        await axn.addDepositReserveLiquidityIx();
        return axn;
    }
    static async buildRedeemReserveCollateralTxns(connection, amount, symbol, publicKey, environment = "production", lendingMarketAddress) {
        const axn = await RelendAction.initialize("redeem", amount, symbol, publicKey, connection, environment, lendingMarketAddress);
        await axn.addSupportIxs("redeem");
        await axn.addRedeemReserveCollateralIx();
        return axn;
    }
    static async buildDepositObligationCollateralTxns(connection, amount, symbol, publicKey, environment = "production", lendingMarketAddress) {
        const axn = await RelendAction.initialize("depositCollateral", amount, symbol, publicKey, connection, environment, lendingMarketAddress);
        await axn.addSupportIxs("depositCollateral");
        await axn.addDepositObligationCollateralIx();
        return axn;
    }
    static async buildWithdrawTxns(connection, amount, symbol, publicKey, environment = "production", lendingMarketAddress) {
        const axn = await RelendAction.initialize("withdraw", amount, symbol, publicKey, connection, environment, lendingMarketAddress);
        await axn.addSupportIxs("withdraw");
        await axn.addWithdrawIx();
        return axn;
    }
    static async buildRepayTxns(connection, amount, symbol, publicKey, environment = "production", lendingMarketAddress) {
        const axn = await RelendAction.initialize("repay", amount, symbol, publicKey, connection, environment, lendingMarketAddress);
        await axn.addSupportIxs("repay");
        await axn.addRepayIx();
        return axn;
    }
    async getTransactions() {
        const txns = {
            preLendingTxn: null,
            lendingTxn: null,
            postLendingTxn: null,
        };
        if (this.preTxnIxs.length) {
            txns.preLendingTxn = new web3_js_1.Transaction({
                feePayer: this.publicKey,
                recentBlockhash: (await this.connection.getRecentBlockhash()).blockhash,
            }).add(...this.preTxnIxs);
        }
        txns.lendingTxn = new web3_js_1.Transaction({
            feePayer: this.publicKey,
            recentBlockhash: (await this.connection.getRecentBlockhash()).blockhash,
        }).add(...this.setupIxs, ...this.lendingIxs, ...this.cleanupIxs);
        if (this.postTxnIxs.length) {
            txns.postLendingTxn = new web3_js_1.Transaction({
                feePayer: this.publicKey,
                recentBlockhash: (await this.connection.getRecentBlockhash()).blockhash,
            }).add(...this.postTxnIxs);
        }
        return txns;
    }
    async sendTransactions(sendTransaction, preCallback, lendingCallback, postCallback) {
        const txns = await this.getTransactions();
        await this.sendSingleTransaction(txns.preLendingTxn, sendTransaction, preCallback);
        const signature = await this.sendSingleTransaction(txns.lendingTxn, sendTransaction, lendingCallback);
        await this.sendSingleTransaction(txns.postLendingTxn, sendTransaction, postCallback);
        return signature;
    }
    async sendSingleTransaction(txn, sendTransaction, callback) {
        if (!txn)
            return "";
        const signature = await sendTransaction(txn, this.connection);
        if (callback) {
            callback();
        }
        await this.connection.confirmTransaction(signature);
        return signature;
    }
    addDepositIx() {
        this.lendingIxs.push((0, instructions_1.depositReserveLiquidityAndObligationCollateralInstruction)(this.amount, this.userTokenAccountAddress, this.userCollateralAccountAddress, new web3_js_1.PublicKey(this.reserve.address), new web3_js_1.PublicKey(this.reserve.liquidityAddress), new web3_js_1.PublicKey(this.reserve.collateralMintAddress), new web3_js_1.PublicKey(this.lendingMarket.address), new web3_js_1.PublicKey(this.lendingMarket.authorityAddress), new web3_js_1.PublicKey(this.reserve.collateralSupplyAddress), // destinationCollateral
        this.obligationAddress, // obligation
        this.publicKey, // obligationOwner
        new web3_js_1.PublicKey(this.reserve.pythOracle), new web3_js_1.PublicKey(this.reserve.switchboardOracle), this.publicKey, // transferAuthority
        this.programId));
    }
    addDepositReserveLiquidityIx() {
        this.lendingIxs.push((0, instructions_1.depositReserveLiquidityInstruction)(this.amount, this.userTokenAccountAddress, this.userCollateralAccountAddress, new web3_js_1.PublicKey(this.reserve.address), new web3_js_1.PublicKey(this.reserve.liquidityAddress), new web3_js_1.PublicKey(this.reserve.collateralMintAddress), new web3_js_1.PublicKey(this.lendingMarket.address), new web3_js_1.PublicKey(this.lendingMarket.authorityAddress), this.publicKey, // transferAuthority
        this.programId));
    }
    addRedeemReserveCollateralIx() {
        this.lendingIxs.push((0, instructions_1.redeemReserveCollateralInstruction)(this.amount, this.userCollateralAccountAddress, this.userTokenAccountAddress, new web3_js_1.PublicKey(this.reserve.address), new web3_js_1.PublicKey(this.reserve.collateralMintAddress), new web3_js_1.PublicKey(this.reserve.liquidityAddress), new web3_js_1.PublicKey(this.lendingMarket.address), // lendingMarket
        new web3_js_1.PublicKey(this.lendingMarket.authorityAddress), // lendingMarketAuthority
        this.publicKey, // transferAuthority
        this.programId));
    }
    addDepositObligationCollateralIx() {
        this.lendingIxs.push((0, instructions_1.depositObligationCollateralInstruction)(this.amount, this.userCollateralAccountAddress, new web3_js_1.PublicKey(this.reserve.collateralSupplyAddress), new web3_js_1.PublicKey(this.reserve.address), this.obligationAddress, // obligation
        new web3_js_1.PublicKey(this.lendingMarket.address), this.publicKey, // obligationOwner
        this.publicKey, // transferAuthority
        this.programId));
    }
    addBorrowIx() {
        this.lendingIxs.push((0, instructions_1.borrowObligationLiquidityInstruction)(this.amount, new web3_js_1.PublicKey(this.reserve.liquidityAddress), this.userTokenAccountAddress, new web3_js_1.PublicKey(this.reserve.address), new web3_js_1.PublicKey(this.reserve.liquidityFeeReceiverAddress), this.obligationAddress, new web3_js_1.PublicKey(this.lendingMarket.address), // lendingMarket
        new web3_js_1.PublicKey(this.lendingMarket.authorityAddress), // lendingMarketAuthority
        this.publicKey, this.programId, this.hostAta));
    }
    async addWithdrawIx() {
        const buffer = await this.connection.getAccountInfo(new web3_js_1.PublicKey(this.reserve.address), "processed");
        if (!buffer) {
            throw Error(`Unable to fetch reserve data for ${this.reserve.liquidityToken.name}`);
        }
        const parsedData = (0, reserve_1.parseReserve)(new web3_js_1.PublicKey(this.reserve.address), buffer)?.info;
        if (!parsedData) {
            throw Error(`Unable to parse data of reserve ${this.reserve.liquidityToken.name}`);
        }
        if (!this.obligationAccountInfo) {
            throw Error(`No Obligation Account Info`);
        }
        // Accure Interest
        const reserveData = this.formatReserveData(parsedData);
        const currentBorrowRate = this.calculateBorrowAPR(parsedData);
        const currentSlot = await this.connection.getSlot();
        const slotElapsed = currentSlot - (parsedData.lastUpdate.slot.toNumber() ?? 0) + WITHDRAW_BUFFER_SLOTS;
        const compoundedInterestRateWhenRefreshReserve = this.calculateCompoundedInterestRate(currentBorrowRate, slotElapsed);
        const reserveBorrowedAmount = new bignumber_js_1.default(parsedData.liquidity.borrowedAmountWads.toString()).div(constants_2.WAD);
        const netNewDebt = reserveBorrowedAmount
            .multipliedBy(compoundedInterestRateWhenRefreshReserve)
            .minus(reserveBorrowedAmount);
        const newBorrowedAmountAfterRefresh = reserveBorrowedAmount.plus(netNewDebt);
        const accumulatedProtocolFees = new bignumber_js_1.default(parsedData.liquidity.accumulatedProtocolFeesWads.toString()).div(constants_2.WAD);
        const newAccumulatedProtocolTakeFee = netNewDebt
            .multipliedBy(reserveData.protocolTakeRate)
            .plus(accumulatedProtocolFees);
        const availableAmount = new bignumber_js_1.default(parsedData.liquidity.availableAmount.toString());
        const totalLiquidity = availableAmount.plus(newBorrowedAmountAfterRefresh).minus(newAccumulatedProtocolTakeFee);
        const mintTotalSupply = parsedData.collateral.mintTotalSupply;
        const cTokenExchangeRate = new bignumber_js_1.default(totalLiquidity.div(new bignumber_js_1.default(mintTotalSupply.toString())));
        this.lendingIxs.push((0, instructions_1.withdrawObligationCollateralAndRedeemReserveLiquidity)(this.amount.eq(new bn_js_1.default(constants_2.U64_MAX))
            ? this.amount
            : new bn_js_1.default(new bignumber_js_1.default(this.amount.toString())
                .dividedBy(cTokenExchangeRate)
                .integerValue(bignumber_js_1.default.ROUND_HALF_FLOOR)
                .toString()), new web3_js_1.PublicKey(this.reserve.collateralSupplyAddress), this.userCollateralAccountAddress, new web3_js_1.PublicKey(this.reserve.address), this.obligationAddress, new web3_js_1.PublicKey(this.lendingMarket.address), new web3_js_1.PublicKey(this.lendingMarket.authorityAddress), this.userTokenAccountAddress, // destinationLiquidity
        new web3_js_1.PublicKey(this.reserve.collateralMintAddress), new web3_js_1.PublicKey(this.reserve.liquidityAddress), this.publicKey, // obligationOwner
        this.publicKey, // transferAuthority
        this.programId));
    }
    async addRepayIx() {
        this.lendingIxs.push((0, instructions_1.repayObligationLiquidityInstruction)(this.amount, this.userTokenAccountAddress, new web3_js_1.PublicKey(this.reserve.liquidityAddress), new web3_js_1.PublicKey(this.reserve.address), this.obligationAddress, new web3_js_1.PublicKey(this.lendingMarket.address), this.publicKey, this.programId));
    }
    async addSupportIxs(action) {
        if (["withdraw", "borrow"].includes(action)) {
            await this.addRefreshIxs();
        }
        if (!["mint", "redeem"].includes(action)) {
            await this.addObligationIxs();
        }
        await this.addAtaIxs(action);
    }
    async addRefreshIxs() {
        // Union of addresses
        const allReserveAddresses = [
            ...Array.from([
                ...this.depositReserves.map((e) => e.toBase58()),
                ...this.borrowReserves.map((e) => e.toBase58()),
                this.reserve.address,
            ]),
        ];
        const uniqueAllReserveAddresses = [...new Set(allReserveAddresses)];
        const pda = (0, constants_1.getPDA)();
        uniqueAllReserveAddresses.forEach((reserveAddress) => {
            const reserveInfo = this.lendingMarket.reserves.find((reserve) => reserve.address === reserveAddress);
            if (!reserveInfo) {
                throw new Error(`Could not find asset ${reserveAddress} in reserves`);
            }
            const symbol = reserveInfo.liquidityToken.symbol;
            let pair;
            if (symbol === constants_1.RE_USD) {
                pair = REUSD_PRICE;
            }
            else if (symbol === constants_1.RE_VND) {
                pair = REVND_PRICE;
            }
            else if (symbol === constants_1.RE_NGN) {
                pair = RENGN_PRICE;
            }
            else {
                pair = PRICE_PAIRS.find((pair) => pair.baseSymbol === reserveInfo.liquidityToken.symbol);
            }
            const oracleProduct = pda.product(new web3_js_1.PublicKey(pair.quoteToken), new web3_js_1.PublicKey(pair.baseToken));
            const refreshReserveIx = (0, instructions_1.refreshReserveInstruction)(new web3_js_1.PublicKey(reserveAddress), this.programId, oracleProduct.key, new web3_js_1.PublicKey(reserveInfo.pythOracle), new web3_js_1.PublicKey(reserveInfo.switchboardOracle));
            this.setupIxs.push(refreshReserveIx);
        });
        const refreshObligationIx = (0, instructions_1.refreshObligationInstruction)(this.obligationAddress, this.depositReserves, this.borrowReserves, this.programId);
        this.setupIxs.push(refreshObligationIx);
    }
    async addObligationIxs() {
        if (!this.obligationAccountInfo) {
            const obligationAccountInfoRentExempt = await this.connection.getMinimumBalanceForRentExemption(obligation_1.OBLIGATION_SIZE);
            this.setupIxs.push(web3_js_1.SystemProgram.createAccountWithSeed({
                fromPubkey: this.publicKey,
                newAccountPubkey: this.obligationAddress,
                basePubkey: this.publicKey,
                seed: this.seed,
                lamports: obligationAccountInfoRentExempt,
                space: obligation_1.OBLIGATION_SIZE,
                programId: this.programId,
            }));
            const initObligationIx = (0, instructions_1.initObligationInstruction)(this.obligationAddress, new web3_js_1.PublicKey(this.lendingMarket.address), this.publicKey, this.programId);
            this.setupIxs.push(initObligationIx);
        }
    }
    async addAtaIxs(action) {
        if (this.symbol === "RENEC") {
            await this.updateWRENECAccount(action);
        }
        if ((action === "withdraw" || action === "borrow" || action === "redeem") && this.symbol !== "RENEC") {
            const userTokenAccountInfo = await this.connection.getAccountInfo(this.userTokenAccountAddress);
            if (!userTokenAccountInfo) {
                const createUserTokenAccountIx = (0, utils_1.createAssociatedTokenAccountInstruction)(this.publicKey, this.userTokenAccountAddress, this.publicKey, new web3_js_1.PublicKey(this.reserve.liquidityToken.mint));
                if (this.positions === constants_1.POSITION_LIMIT && this.hostAta) {
                    this.preTxnIxs.push(createUserTokenAccountIx);
                }
                else {
                    this.setupIxs.push(createUserTokenAccountIx);
                }
            }
        }
        if (action === "withdraw" || action === "mint" || action === "deposit") {
            const userCollateralAccountInfo = await this.connection.getAccountInfo(this.userCollateralAccountAddress);
            if (!userCollateralAccountInfo) {
                const createUserCollateralAccountIx = (0, utils_1.createAssociatedTokenAccountInstruction)(this.publicKey, this.userCollateralAccountAddress, this.publicKey, new web3_js_1.PublicKey(this.reserve.collateralMintAddress));
                if (this.positions === constants_1.POSITION_LIMIT && this.symbol === "RENEC") {
                    this.preTxnIxs.push(createUserCollateralAccountIx);
                }
                else {
                    this.setupIxs.push(createUserCollateralAccountIx);
                }
            }
        }
    }
    async updateWRENECAccount(action) {
        const preIxs = [];
        const postIxs = [];
        let safeRepay = new bn_js_1.default(this.amount);
        if (this.obligationAccountInfo && action === "repay" && this.amount.eq(new bn_js_1.default(constants_2.U64_MAX))) {
            const buffer = await this.connection.getAccountInfo(new web3_js_1.PublicKey(this.reserve.address), "processed");
            if (!buffer) {
                throw Error(`Unable to fetch reserve data for ${this.reserve.liquidityToken.name}`);
            }
            const parsedData = (0, reserve_1.parseReserve)(new web3_js_1.PublicKey(this.reserve.address), buffer)?.info;
            if (!parsedData) {
                throw Error(`Unable to parse data of reserve ${this.reserve.liquidityToken.name}`);
            }
            const borrow = this.obligationAccountInfo.borrows.find((borrow) => borrow.borrowReserve.toBase58() === this.reserve.address);
            if (!borrow) {
                throw Error(`Unable to find obligation borrow to repay for ${this.obligationAccountInfo.owner.toBase58()}`);
            }
            safeRepay = new bn_js_1.default(Math.floor(new bignumber_js_1.default(borrow.borrowedAmountWads.toString())
                .multipliedBy(parsedData.liquidity.cumulativeBorrowRateWads.toString())
                .dividedBy(borrow.cumulativeBorrowRateWads.toString())
                .dividedBy(constants_2.WAD)
                .plus(SOL_PADDING_FOR_INTEREST)
                .toNumber()).toString());
        }
        const userWRENECAccountInfo = await this.connection.getAccountInfo(this.userTokenAccountAddress);
        const rentExempt = await spl_token_1.Token.getMinBalanceRentForExemptAccount(this.connection);
        const sendAction = action === "deposit" || action === "repay" || action === "mint";
        const transferLamportsIx = web3_js_1.SystemProgram.transfer({
            fromPubkey: this.publicKey,
            toPubkey: this.userTokenAccountAddress,
            lamports: (userWRENECAccountInfo ? 0 : rentExempt) + (sendAction ? parseInt(safeRepay.toString(), 10) : 0),
        });
        preIxs.push(transferLamportsIx);
        const closeWRENECAccountIx = spl_token_1.Token.createCloseAccountInstruction(spl_token_1.TOKEN_PROGRAM_ID, this.userTokenAccountAddress, this.publicKey, this.publicKey, []);
        if (userWRENECAccountInfo) {
            const syncIx = (0, instructions_1.syncNative)(this.userTokenAccountAddress);
            if (sendAction) {
                preIxs.push(syncIx);
            }
            else {
                postIxs.push(closeWRENECAccountIx);
            }
        }
        else {
            const createUserWRENECAccountIx = (0, utils_1.createAssociatedTokenAccountInstruction)(this.publicKey, this.userTokenAccountAddress, this.publicKey, spl_token_1.NATIVE_MINT);
            preIxs.push(createUserWRENECAccountIx);
            postIxs.push(closeWRENECAccountIx);
        }
        if (this.positions && this.positions >= constants_1.POSITION_LIMIT) {
            this.preTxnIxs.push(...preIxs);
            this.postTxnIxs.push(...postIxs);
        }
        else {
            this.setupIxs.push(...preIxs);
            this.cleanupIxs.push(...postIxs);
        }
    }
    formatReserveData(parsedData) {
        const totalBorrowsWads = parsedData.liquidity.borrowedAmountWads;
        const totalLiquidityWads = parsedData.liquidity.availableAmount.mul(new bn_js_1.default(constants_2.WAD));
        const totalDepositsWads = totalBorrowsWads.add(totalLiquidityWads);
        const cTokenExchangeRate = new bignumber_js_1.default(totalDepositsWads.toString())
            .div(parsedData.collateral.mintTotalSupply.toString())
            .div(constants_2.WAD)
            .toNumber();
        return {
            // Reserve config
            optimalUtilizationRate: parsedData.config.optimalUtilizationRate / 100,
            loanToValueRatio: parsedData.config.loanToValueRatio / 100,
            liquidationBonus: parsedData.config.liquidationBonus / 100,
            liquidationThreshold: parsedData.config.liquidationThreshold / 100,
            minBorrowRate: parsedData.config.minBorrowRate / 100,
            optimalBorrowRate: parsedData.config.optimalBorrowRate / 100,
            maxBorrowRate: parsedData.config.maxBorrowRate / 100,
            protocolTakeRate: parsedData.config.protocolTakeRate / 100,
            borrowFeePercentage: new bignumber_js_1.default(parsedData.config.fees.borrowFeeWad.toString()).dividedBy(constants_2.WAD).toNumber(),
            hostFeePercentage: parsedData.config.fees.hostFeePercentage / 100,
            flashLoanFeePercentage: new bignumber_js_1.default(parsedData.config.fees.flashLoanFeeWad.toString())
                .dividedBy(constants_2.WAD)
                .toNumber(),
            depositLimit: parsedData.config.depositLimit,
            reserveBorrowLimit: parsedData.config.borrowLimit,
            // Reserve info
            name: this.reserve.liquidityToken.name,
            symbol: this.reserve.liquidityToken.symbol,
            decimals: this.reserve.liquidityToken.decimals,
            mintAddress: this.reserve.liquidityToken.mint,
            totalDepositsWads,
            totalBorrowsWads,
            totalLiquidityWads,
            supplyInterestAPY: this.calculateSupplyAPY(parsedData),
            borrowInterestAPY: this.calculateBorrowAPY(parsedData),
            assetPriceUSD: new bignumber_js_1.default(parsedData.liquidity.marketPrice.toString()).div(constants_2.WAD).toNumber(),
            userDepositLimit: this.reserve.userSupplyCap,
            cumulativeBorrowRateWads: parsedData.liquidity.cumulativeBorrowRateWads,
            cTokenExchangeRate,
        };
    }
    calculateSupplyAPY = (reserve) => {
        const apr = this.calculateSupplyAPR(reserve);
        const apy = new bignumber_js_1.default(1).plus(new bignumber_js_1.default(apr).dividedBy(constants_1.SLOTS_PER_YEAR)).toNumber() ** constants_1.SLOTS_PER_YEAR - 1;
        return apy;
    };
    calculateBorrowAPY = (reserve) => {
        const apr = this.calculateBorrowAPR(reserve);
        const apy = new bignumber_js_1.default(1).plus(new bignumber_js_1.default(apr).dividedBy(constants_1.SLOTS_PER_YEAR)).toNumber() ** constants_1.SLOTS_PER_YEAR - 1;
        return apy;
    };
    calculateSupplyAPR(reserve) {
        const currentUtilization = this.calculateUtilizationRatio(reserve);
        const borrowAPY = this.calculateBorrowAPR(reserve);
        return currentUtilization * borrowAPY;
    }
    calculateUtilizationRatio(reserve) {
        const totalBorrowsWads = new bignumber_js_1.default(reserve.liquidity.borrowedAmountWads.toString()).div(constants_2.WAD);
        const currentUtilization = totalBorrowsWads
            .dividedBy(totalBorrowsWads.plus(reserve.liquidity.availableAmount.toString()))
            .toNumber();
        return currentUtilization;
    }
    calculateBorrowAPR(reserve) {
        const currentUtilization = this.calculateUtilizationRatio(reserve);
        const optimalUtilization = reserve.config.optimalUtilizationRate / 100;
        let borrowAPR;
        if (optimalUtilization === 1.0 || currentUtilization < optimalUtilization) {
            const normalizedFactor = currentUtilization / optimalUtilization;
            const optimalBorrowRate = reserve.config.optimalBorrowRate / 100;
            const minBorrowRate = reserve.config.minBorrowRate / 100;
            borrowAPR = normalizedFactor * (optimalBorrowRate - minBorrowRate) + minBorrowRate;
        }
        else {
            const normalizedFactor = (currentUtilization - optimalUtilization) / (1 - optimalUtilization);
            const optimalBorrowRate = reserve.config.optimalBorrowRate / 100;
            const maxBorrowRate = reserve.config.maxBorrowRate / 100;
            borrowAPR = normalizedFactor * (maxBorrowRate - optimalBorrowRate) + optimalBorrowRate;
        }
        return borrowAPR;
    }
    calculateCompoundedInterestRate(currentBorrowRate, slotElapsed) {
        const slotInterestRate = new bignumber_js_1.default(currentBorrowRate).div(constants_1.SLOTS_PER_YEAR);
        const compoundInterestRate = (1 + slotInterestRate.toNumber()) ** slotElapsed;
        return compoundInterestRate;
    }
}
exports.RelendAction = RelendAction;
