"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RelendMarket = void 0;
const web3_js_1 = require("@solana/web3.js");
const axios_1 = __importDefault(require("axios"));
const core_1 = require("../core");
const state_1 = require("../state");
const constants_1 = require("./constants");
const obligation_1 = require("./obligation");
const reserve_1 = require("./reserve");
// const API_ENDPOINT = "https://api.solend.fi"
const API_ENDPOINT = process.env.NEXT_PUBLIC_BACKEND_URL;
class RelendMarket {
    connection;
    reserves;
    rewardsData;
    config;
    programId;
    constructor(connection, config, reserves, programId) {
        this.connection = connection;
        this.reserves = reserves;
        this.rewardsData = null;
        this.config = config;
        this.programId = programId;
    }
    static async initialize(connection, environment = "production", marketAddress) {
        // const config = (await (
        //   await axios.get(
        //     `${API_ENDPOINT}/v1/markets/configs?scope=all&deployment=${environment}`
        //   )
        // ).data) as Config;
        const config = (0, core_1.getRelendInfo)();
        let marketConfig;
        if (marketAddress) {
            marketConfig = config.find((market) => market.address == marketAddress) ?? null;
            if (!marketConfig) {
                throw `market address not found: ${marketAddress}`;
            }
        }
        else {
            marketConfig = config.find((market) => market.isPrimary) ?? config[0];
        }
        const reserves = marketConfig.reserves.map((res) => new reserve_1.RelendReserve(res, connection));
        return new RelendMarket(connection, marketConfig, reserves, (0, constants_1.getProgramId)());
    }
    async fetchObligationByWallet(publicKey) {
        const { config, reserves } = this;
        if (!config) {
            throw Error("Market must be initialized to call initialize.");
        }
        const obligationAddress = await web3_js_1.PublicKey.createWithSeed(publicKey, config.address.slice(0, 32), this.programId);
        const rawObligationData = await this.connection.getAccountInfo(obligationAddress);
        if (!rawObligationData) {
            return null;
        }
        const parsedObligation = (0, state_1.parseObligation)(web3_js_1.PublicKey.default, rawObligationData);
        if (!parsedObligation) {
            throw Error("Could not parse obligation.");
        }
        if (!reserves.every((reserve) => reserve.stats)) {
            await this.loadReserves();
        }
        const obligationInfo = parsedObligation.info;
        return new obligation_1.RelendObligation(publicKey, obligationAddress, obligationInfo, reserves);
    }
    async loadAll() {
        const promises = [this.loadReserves(), this.loadRewards()];
        await Promise.all(promises);
    }
    async loadExternalRewardData() {
        const data = (await axios_1.default.get(`${API_ENDPOINT}/liquidity-mining/external-reward-stats-v2?flat=true`))
            .data;
        return data;
    }
    async loadPriceData(symbols) {
        const data = (await (await axios_1.default.get(`${API_ENDPOINT}/v1/prices/?symbols=${symbols.join(",")}`)).data);
        return data.results.reduce((acc, price) => ({
            ...acc,
            [price.identifier]: Number(price.price),
        }), {});
    }
    getLatestRewardRate(rewardRates, slot) {
        return rewardRates
            .filter((rr) => slot >= rr.beginningSlot)
            .reduce((v1, v2) => (v1.beginningSlot > v2.beginningSlot ? v1 : v2), {
            beginningSlot: 0,
            rewardRate: "0",
        });
    }
    async loadRewards() {
        if (!this.config) {
            throw Error("Market must be initialized to call loadRewards.");
        }
        const promises = [this.loadExternalRewardData(), this.connection.getSlot("finalized")];
        const [externalRewards, currentSlot] = await Promise.all(promises);
        const querySymbols = [...Array.from(new Set(externalRewards.map((reward) => reward.rewardSymbol)))];
        const priceData = await this.loadPriceData(querySymbols.concat("SLND"));
        this.rewardsData = this.reserves.reduce((acc, reserve) => {
            const supply = [
                ...externalRewards
                    .filter((externalReward) => externalReward.reserveID === reserve.config.address && externalReward.side === "supply")
                    .map((externalReward) => ({
                    rewardRate: this.getLatestRewardRate(externalReward.rewardRates, currentSlot).rewardRate,
                    rewardMint: externalReward.rewardMint,
                    rewardSymbol: externalReward.rewardSymbol,
                    price: priceData[externalReward.rewardSymbol],
                })),
            ].filter(Boolean);
            const borrow = [
                ...externalRewards
                    .filter((externalReward) => externalReward.reserveID === reserve.config.address && externalReward.side === "borrow")
                    .map((externalReward) => ({
                    rewardRate: this.getLatestRewardRate(externalReward.rewardRates, currentSlot).rewardRate,
                    rewardMint: externalReward.rewardMint,
                    rewardSymbol: externalReward.rewardSymbol,
                    price: priceData[externalReward.rewardSymbol],
                })),
            ].filter(Boolean);
            return {
                ...acc,
                [reserve.config.liquidityToken.mint]: {
                    supply,
                    borrow,
                },
            };
        }, {});
        const refreshReserves = this.reserves.map((reserve) => {
            return reserve.load(this.rewardsData ?? undefined);
        });
        await Promise.all(refreshReserves);
    }
    async loadReserves() {
        const addresses = this.reserves.map((reserve) => new web3_js_1.PublicKey(reserve.config.address));
        const reserveAccounts = await this.connection.getMultipleAccountsInfo(addresses, "processed");
        const loadReserves = this.reserves.map((reserve, index) => {
            reserve.setBuffer(reserveAccounts[index]);
            return reserve.load();
        });
        await Promise.all(loadReserves);
    }
    async refreshAll() {
        const promises = [
            this.reserves.every((reserve) => reserve.stats) ? this.loadReserves() : null,
            this.rewardsData ? this.loadRewards() : null,
        ].filter((x) => x);
        await Promise.all(promises);
    }
}
exports.RelendMarket = RelendMarket;
