"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.fetchObligationsByAddress = exports.fetchObligationByAddress = exports.formatObligation = void 0;
const web3_js_1 = require("@solana/web3.js");
const bignumber_js_1 = __importDefault(require("bignumber.js"));
const bn_js_1 = require("bn.js");
const constants_1 = require("../../classes/constants");
const state_1 = require("../../state");
const constants_2 = require("../constants");
const utils_1 = require("./utils");
function formatObligation(obligation, pool) {
    const poolAddress = obligation.info.lendingMarket.toBase58();
    let minPriceUserTotalSupply = new bignumber_js_1.default(0);
    let minPriceBorrowLimit = new bignumber_js_1.default(0);
    let maxPriceUserTotalWeightedBorrow = new bignumber_js_1.default(0);
    const deposits = obligation.info.deposits
        .filter((d) => d.depositedAmount.toString() !== "0")
        .map((d) => {
        const reserveAddress = d.depositReserve.toBase58();
        const reserve = pool.reserves.find((r) => r.address === reserveAddress);
        if (constants_2.IS_DEBUGGING) {
            console.log("formatObligation");
        }
        if (!reserve)
            throw Error("Deposit in obligation does not exist in the pool");
        const amount = new bignumber_js_1.default(d.depositedAmount.toString())
            .shiftedBy(-reserve.decimals)
            .times(reserve.cTokenExchangeRate);
        const depositedAmount = new bignumber_js_1.default(d.depositedAmount.toString()).shiftedBy(-reserve.decimals);
        // Calculate the earned amount
        const totalLiquidityWads = reserve.availableAmountWad.mul(new bn_js_1.BN(constants_1.WAD));
        const totalDepositsWads = reserve.totalBorrowWad.add(totalLiquidityWads);
        const cTokenExchangeRate = totalDepositsWads.div(reserve.mintTotalSupplyWad).div(new bn_js_1.BN(constants_1.WAD));
        const earnedAmount = d.depositedAmount.mul(cTokenExchangeRate).sub(d.depositedAmount);
        const amountUsd = amount.times(reserve.price);
        minPriceUserTotalSupply = minPriceUserTotalSupply.plus(amount.times(reserve.minPrice));
        minPriceBorrowLimit = minPriceBorrowLimit.plus(amount.times(reserve.minPrice).times(reserve.loanToValueRatio));
        return {
            liquidationThreshold: reserve.liquidationThreshold,
            loanToValueRatio: reserve.loanToValueRatio,
            symbol: reserve.symbol,
            price: reserve.price,
            reserveAddress,
            amount,
            amountUsd,
            depositedAmount,
            earnedAmount: new bignumber_js_1.default(earnedAmount.toString()),
        };
    });
    const borrows = obligation.info.borrows
        .filter((b) => b.borrowedAmountWads.toString() !== "0")
        .map((b) => {
        const reserveAddress = b.borrowReserve.toBase58();
        const reserve = pool.reserves.find((r) => r.address === reserveAddress);
        if (!reserve)
            throw Error("Borrow in obligation does not exist in the pool");
        const amount = new bignumber_js_1.default(b.borrowedAmountWads.toString())
            .shiftedBy(-18 - reserve.decimals)
            .times(reserve.cumulativeBorrowRate)
            .dividedBy(new bignumber_js_1.default(b.cumulativeBorrowRateWads.toString()).shiftedBy(-18));
        const amountUsd = amount.times(reserve.price);
        const maxPrice = reserve.emaPrice ? bignumber_js_1.default.max(reserve.emaPrice, reserve.price) : reserve.price;
        maxPriceUserTotalWeightedBorrow = maxPriceUserTotalWeightedBorrow.plus(amount.times(maxPrice).times(reserve.borrowWeight ? reserve.borrowWeight : constants_1.U64_MAX));
        const cumulativeBorrowRate = new bignumber_js_1.default(b.cumulativeBorrowRateWads.toString()).shiftedBy(-18);
        return {
            liquidationThreshold: reserve.liquidationThreshold,
            loanToValueRatio: reserve.loanToValueRatio,
            symbol: reserve.symbol,
            price: reserve.price,
            cumulativeBorrowRate,
            reserveAddress,
            amount,
            amountUsd,
            weightedAmountUsd: new bignumber_js_1.default(reserve.borrowWeight).multipliedBy(amountUsd),
        };
    });
    const totalSupplyValue = deposits.reduce((acc, d) => acc.plus(d.amountUsd), new bignumber_js_1.default(0));
    const totalBorrowValue = borrows.reduce((acc, b) => acc.plus(b.amountUsd), new bignumber_js_1.default(0));
    const weightedTotalBorrowValue = borrows.reduce((acc, b) => acc.plus(b.weightedAmountUsd), new bignumber_js_1.default(0));
    const borrowLimit = deposits.reduce((acc, d) => d.amountUsd.times(d.loanToValueRatio).plus(acc), (0, bignumber_js_1.default)(0));
    const liquidationThreshold = deposits.reduce((acc, d) => d.amountUsd.times(d.liquidationThreshold).plus(acc), (0, bignumber_js_1.default)(0));
    const netAccountValue = totalSupplyValue.minus(totalBorrowValue);
    const liquidationThresholdFactor = totalSupplyValue.isZero()
        ? new bignumber_js_1.default(0)
        : liquidationThreshold.dividedBy(totalSupplyValue);
    const borrowLimitFactor = totalSupplyValue.isZero() ? new bignumber_js_1.default(0) : borrowLimit.dividedBy(totalSupplyValue);
    const borrowUtilization = borrowLimit.isZero() ? new bignumber_js_1.default(0) : totalBorrowValue.dividedBy(borrowLimit);
    const weightedBorrowUtilization = minPriceBorrowLimit.isZero()
        ? new bignumber_js_1.default(0)
        : weightedTotalBorrowValue.dividedBy(borrowLimit);
    const isBorrowLimitReached = borrowUtilization.isGreaterThanOrEqualTo(new bignumber_js_1.default("1"));
    const borrowOverSupply = totalSupplyValue.isZero() ? new bignumber_js_1.default(0) : totalBorrowValue.dividedBy(totalSupplyValue);
    const positions = obligation.info.deposits.filter((d) => !d.depositedAmount.isZero()).length +
        obligation.info.borrows.filter((b) => !b.borrowedAmountWads.isZero()).length;
    const weightedConservativeBorrowUtilization = minPriceBorrowLimit.isZero()
        ? new bignumber_js_1.default(0)
        : maxPriceUserTotalWeightedBorrow.dividedBy(minPriceBorrowLimit);
    return {
        address: obligation.pubkey.toBase58(),
        positions,
        deposits,
        borrows,
        poolAddress,
        totalSupplyValue,
        totalBorrowValue,
        borrowLimit,
        liquidationThreshold,
        netAccountValue,
        liquidationThresholdFactor,
        borrowLimitFactor,
        borrowUtilization,
        weightedConservativeBorrowUtilization,
        weightedBorrowUtilization,
        isBorrowLimitReached,
        borrowOverSupply,
        weightedTotalBorrowValue,
        minPriceUserTotalSupply,
        minPriceBorrowLimit,
        maxPriceUserTotalWeightedBorrow,
    };
}
exports.formatObligation = formatObligation;
async function fetchObligationByAddress(obligationAddress, connection, debug) {
    if (debug)
        console.log("fetchObligationByAddress");
    const rawObligationData = await connection.getAccountInfo(new web3_js_1.PublicKey(obligationAddress));
    if (!rawObligationData) {
        return null;
    }
    const parsedObligation = (0, state_1.parseObligation)(new web3_js_1.PublicKey(obligationAddress), rawObligationData);
    if (!parsedObligation) {
        return null;
    }
    return parsedObligation;
}
exports.fetchObligationByAddress = fetchObligationByAddress;
async function fetchObligationsByAddress(obligationAddresses, connection, debug) {
    if (debug)
        console.log("fetchObligationsByAddress", obligationAddresses.length);
    const rawObligations = await (0, utils_1.getBatchMultipleAccountsInfo)(obligationAddresses, connection);
    const parsedObligations = rawObligations
        .map((obligation, index) => obligation ? (0, state_1.parseObligation)(new web3_js_1.PublicKey(obligationAddresses[index]), obligation) : null)
        .filter(Boolean);
    return parsedObligations;
}
exports.fetchObligationsByAddress = fetchObligationsByAddress;
