"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.fetchPrices = void 0;
const web3_js_1 = require("@solana/web3.js");
const custom_oracle_1 = require("../../utils/custom-oracle");
const constants_1 = require("../constants");
const DEFAULT_PRICES = {
    reBTC: 26_262.76,
    reETH: 1_588.81,
    RENEC: 0.3874,
    reVND: 0.000040760609273283,
    reNGN: 0.08,
};
const REUSD_PRICE = (0, constants_1.getReUSDPricePair)();
const PRICE_PAIRS = (0, constants_1.getPricePair)();
async function fetchPrices(parsedReserves, connection, debug) {
    if (debug)
        console.log("fetchPrices");
    // const solendInfo = (await (
    //   await axios.get(
    //     `${API_ENDPOINT}/v1/markets/configs?scope=all&deployment=${environment}`
    //   )
    // ).data) as ConfigType;
    const relendInfo = (0, constants_1.getRelendInfo)();
    const pda = (0, constants_1.getPDA)();
    const lendingMarket = relendInfo.find((market) => market.isPrimary) ?? relendInfo[0];
    const pricePairs = parsedReserves.reduce((acc, reserve) => {
        const symbol = lendingMarket.reserves.find((r) => r.address === reserve.pubkey.toBase58())?.liquidityToken.symbol;
        if (!symbol) {
            return acc;
        }
        if (symbol === constants_1.RE_USD) {
            const priceWithAddress = {
                ...REUSD_PRICE,
                reserveAddress: reserve.pubkey.toBase58(),
            };
            acc.push(priceWithAddress);
            return acc;
        }
        let foundPair;
        if (symbol === constants_1.RE_VND || symbol === constants_1.RE_NGN) {
            // because reVND and reNGN is the opposite (reUSD / reVND(reNGN))
            foundPair = PRICE_PAIRS.find((pair) => pair.quoteSymbol === symbol);
        }
        else {
            foundPair = PRICE_PAIRS.find((pair) => pair.baseSymbol === symbol);
        }
        if (!foundPair) {
            return acc;
        }
        const priceWithAddress = {
            ...foundPair,
            reserveAddress: reserve.pubkey.toBase58(),
        };
        acc.push(priceWithAddress);
        return acc;
    }, []);
    const output = {};
    for (const pair of pricePairs) {
        try {
            if (pair.baseSymbol === constants_1.RE_USD && pair.quoteSymbol === constants_1.RE_USD) {
                output[pair.reserveAddress] = {
                    spotPrice: 1,
                    emaPrice: 1,
                };
            }
            else {
                const oracleProduct = pda.product(new web3_js_1.PublicKey(pair.quoteToken), new web3_js_1.PublicKey(pair.baseToken));
                const oraclePrice = pda.price(oracleProduct.key);
                const priceOracle = await (0, custom_oracle_1.parsePriceOracle)(connection, oraclePrice.key);
                const productOracle = await (0, custom_oracle_1.parseProductOracle)(connection, oracleProduct.key);
                const expo = productOracle?.expo ?? 0;
                if (debug) {
                    console.log("pair: ", pair);
                    console.log("expo: ", expo);
                    console.log("oracleProduct key", oracleProduct.key.toBase58());
                    console.log("oraclePrice key", oraclePrice.key.toBase58());
                }
                const defaultPrice = DEFAULT_PRICES[pair.baseSymbol] ?? 1000;
                let spotPrice;
                let emaPrice;
                if (pair.quoteSymbol === constants_1.RE_VND || pair.quoteSymbol === constants_1.RE_NGN) {
                    spotPrice =
                        priceOracle && !priceOracle.price.isZero() && priceOracle.price.toNumber() * Math.pow(10, expo) !== 0
                            ? 1 / (priceOracle.price.toNumber() * Math.pow(10, expo))
                            : defaultPrice;
                    emaPrice =
                        priceOracle && !priceOracle.price.isZero() && priceOracle.price.toNumber() * Math.pow(10, expo) !== 0
                            ? 1 / (priceOracle.price.toNumber() * Math.pow(10, expo))
                            : defaultPrice;
                }
                else {
                    spotPrice =
                        priceOracle && !priceOracle.price.isZero()
                            ? priceOracle.price.toNumber() * Math.pow(10, expo)
                            : defaultPrice;
                    emaPrice =
                        priceOracle && !priceOracle.price.isZero()
                            ? priceOracle.price.toNumber() * Math.pow(10, expo)
                            : defaultPrice;
                }
                output[pair.reserveAddress] = {
                    spotPrice,
                    emaPrice,
                };
            }
        }
        catch (err) {
            console.error("error fetching on pair", pair.baseSymbol, pair.quoteSymbol);
            console.error("error", err);
        }
    }
    if (debug)
        console.log("output", output);
    return output;
}
exports.fetchPrices = fetchPrices;
